# kartgif.pm er en modul med ymse funksjoner for  legge ut
# forskjellig informasjon p eksisterende kart (for eksempel kartene
# generert av [?])
#
# kartgif.pm er basert p "kartgif"-scriptet laget av [...] og Olav
# Kvittem, og objektorienterifisert av Salve J. Nilsen
#
# Eksempel p kjring i bash, for lettere debugging:
#
# $ cd /kart/norgesweb
# $ { perl -Mstrict -wT bin/kartgif -kart geo/grunnlag/norge.ps \
#         -gif geo/grunnlag/norge.gif -koord geo/education.ko \
#         2>&3 | xv - >&- & } 3>&1 | less
#
#
############################################################################
#
# Hvilke viktige variabler finnes?
#
# %{$self->{'punkt'}} - stedsliste over alle institusjoner
# %{$self->{'punktdata'}->{$sted}}} - oversikt over stedsinfo
# %{$self->{'rute'}->{'$rutex'}->{'$rutey'}} - hvilke institusjoener ligger her?
#
# %sone{$x}{$y} - oversikt over hvilke ruter som er okkupert
#
############################################################################


package kartgif;

use lib "/kart/norgesweb/lib/perl";

use GD;

# Modulen GD.pm med dokumentering finnes p adressen
# http://www.genome.wi.mit.edu/pub/software/WWW/GD.html

# Dette programmet benytter GD.pm versjon 1.14

use utmgeo;

sub new {
    my($class) = shift;
    my($defaults) = shift || {};

    my($instance) =
      {
       'ps_kartfil'   => undef,
       'gif_kartfil'  => undef,
       'koordfil'     => undef,
       'last'         => undef,
       'hint'         => undef,
       'blank'        => 0,
       'opt_debug'    => 0,
       'opt_klump'    => 0,
       'opt_n'        => 0,
       'opt_v'        => 0,
       'gif_icon'     => "/kart/norgesweb/lib/icons",
       'defaultLinjeTykkelse' => 3,
       'pkt_storrelse' => 8,
       'sky_storrelse' => 12,
       'textpos'      => undef,
       'textbg'       => undef,

       '_skytype'     => {}, # Interne datastrukturer
       '_linjenavn'   => [],
       '_linjelengde' => {},
       '_sone'        => {},
      };

    foreach my $default (keys %{$defaults}) {
	 if($default =~ m/^_/) {
	     print STDERR "Ignorerer konfigurasjon av $default\n";
	     next;
	 }
	 $instance->{$default} = $defaults->{$default};
    }

    bless $instance, $class;

    return $instance;
}




# Se subrutine tegn_Punkt for def. p punktfarger
# Se subrutine tegn_Linje for def. p linjefarger

# -----------------------------------------------------------
#
# Funksjoner & al.
#
# -----------------------------------------------------------


# pne gif-fil for tegne p
#
# $object = $object->les_Grunnlag_Gif();
#
sub les_Grunnlag_Gif {
    my($self) = shift;
    my $microiconfilename = "gruppe_micro.gif";
    my $pkticonfilename = "gruppe_pkt.gif";
    my $liteniconfilename = "gruppe_liten.gif";
    my $middelsiconfilename = "gruppe_middels.gif";

    # skaff gif

    if ($self->{'blank'}) { # bruk blank bakgrunn
	$self->{'kartgif'} = new GD::Image($self->{'gif_x'}, $self->{'gif_y'});
	my $white = $self->{'kartgif'}->colorAllocate(255,255,255);
	$self->{'kartgif'}->fill(0,0,$white);
    } else { # Hent gif-kart for  tegne p
	$self->{'kartgif'} = $self->les_Gif_icon($self->{'gif_kartfil'});
    }


    $self->{'micro_icon'} = $self->les_Gif_icon($microiconfilename);
    $self->{'pkt_icon'} = $self->les_Gif_icon($pkticonfilename);
    $self->{'liten_icon'} = $self->les_Gif_icon($liteniconfilename);
    $self->{'middels_icon'} = $self->les_Gif_icon($middelsiconfilename);

    $self->{'punktfarger'}->{'lysegraa'} =
	$self->{'kartgif'}->colorAllocate(228,228,228);
    $self->{'punktfarger'}->{'signalrod'} =
	$self->{'kartgif'}->colorAllocate(255,  0,  0);
    $self->{'punktfarger'}->{'signalblaa'} =
	$self->{'kartgif'}->colorAllocate(  0,  0,255);
    $self->{'punktfarger'}->{'rod'} =
	$self->{'kartgif'}->colorAllocate(200,100,100);
    $self->{'punktfarger'}->{'gronn'} =
	$self->{'kartgif'}->colorAllocate(  0,255,  0);
    $self->{'punktfarger'}->{'hvit'} =
	$self->{'kartgif'}->colorAllocate(255,255,255);
    $self->{'punktfarger'}->{'lyseblaa'} =
	$self->{'kartgif'}->colorAllocate(228,228,255);
    $self->{'punktfarger'}->{'lysegul'} =
	$self->{'kartgif'}->colorAllocate(255,255,228);
    $self->{'punktfarger'}->{'blaa'} =
	$self->{'kartgif'}->colorAllocate(100,100,200);
    $self->{'punktfarger'}->{'svart'} =
	$self->{'kartgif'}->colorAllocate(0,0,0);

    return $self;
}

# pne ps-fil for  hente kartkonstanter/-parametre
#
# $object = $object->les_Grunnlag_Ps();
#
sub les_Grunnlag_Ps {
    my($self) = shift;

    my($linje, $konst, $verdi);

    open(PS, $self->{'ps_kartfil'}) ||
	die "Kan ikke pne .ps $self->{'ps_kartfil'}: $!\n";

    while (defined($linje = <PS>)) {
	chomp($linje);
	($konst,$verdi) = split /: /,$linje;
	last if $konst eq "%EndHeader";
      SWITCH: {
	  if ($konst eq "%Skalering") {
	      $self->{'skalering'} = $verdi;
	      last SWITCH;
	  }
	  if ($konst eq "%Giftrans") {
	      $self->{'giftrans'} = $verdi;
	      last SWITCH;
	  }
	  if ($konst eq "%Size") {
	      ($self->{'gif_x'}, $self->{'gif_y'}) = split / /,$verdi;
	      last SWITCH;
	  }
	  if ($konst eq "%Area") {
	      ($self->{'UTM_areal_nord1'},
	       $self->{'UTM_areal_oest1'},
	       $self->{'UTM_areal_nord2'},
	       $self->{'UTM_areal_oest2'}) = split /,/,$verdi;
	      last SWITCH;
	  }
	  if ($konst eq "%Sone") {
	      $self->{'gsone'} = $verdi;
	      last SWITCH;
	  }
      }
    }
    close(PS);
    return $self;
} # slutt les_Grunnlag_Ps

sub les_Gif_icon {
    my $self = shift;
    my $icon_navn = shift;

    if (! defined $self->{'icons'}->{$icon_navn}) {
	open(GIF, "$self->{'gif_icon'}/$icon_navn")
	  || open(GIF,$icon_navn)
	    || die "Kan ikke pne $self->{'gif_icon'}/$icon_navn "
	      ."eller $icon_navn: $!\n";
	$self->{'icons'}->{$icon_navn} = newFromGif GD::Image(*GIF) ||
	  die "$icon_navn har ikke godkjent gif-format: $!";
	close GIF;
    }

    return $self->{'icons'}->{$icon_navn};
}


# -----------------------------------------------------------
#
# Lese av all informasjon i koordinatfil, bl.a. stedsnavn,
# sone, linjetykkelse, farger, UTM-koordinater osv.
#
# Leser ogs inn last-informasjon om nsket.
#
# -----------------------------------------------------------
#
# $object = $object->les_Koordinater()
#
sub les_Koordinater {
    my($self) = shift;

    my($linje, $type);
    my($stedtype, $navn, $sone, $UTM_N, $UTM_E, $pnavn);
    my($xpos, $ypos);
    my($linjetype, $sted1, $sted2, $ptykk);
    my(@farge);

    $linjetype = "linje";

    # Default rutestrrelse
    $self->{'rutesx'} =	$self->{'rutesy'} = $self->{'sky_storrelse'};

    # Text bredde og hyde
    $self->{'tb'} = gdSmallFont->width;
    $self->{'th'} = gdSmallFont->height;

    $self->{'_nudge'} = $self->{'tb'}*2;

    # Rutehyde m minst vre strre eller lik texthyde
    if ($self->{'th'} > $self->{'sky_storrelse'}) {
	$self->{'rutesy'} = $self->{'th'};
    }

    $self->{'ant_x'} = int ($self->{'gif_x'} / $self->{'rutesx'});
    $self->{'ant_y'} = int ($self->{'gif_y'} / $self->{'rutesy'});

    open(KOORD, $self->{'koordfil'}) ||
      die "Kan ikke pne .ko $self->{'koordfil'}: $!\n";

    while (defined($linje = <KOORD>)) {
	next if $linje =~ /^\s*\#|^\s*$/;
	chomp($linje);
	$type = $linje;
	$type =~ s/ .*$//; # Slett alt utenom frste ord

	if ($type eq "PUNKT") {
	    ($type, $stedtype, $navn, $sone, $UTM_N, $UTM_E, $pnavn) =
	      split /\s+/, $linje, 7;
	    if ($sone != $self->{'gsone'}) {
		my $utmgeo = new utmgeo();
		($UTM_N, $UTM_E) =
		  $utmgeo->utmtoutm($UTM_N, $UTM_E, $sone, $self->{'gsone'});
	    }
	    ($xpos,$ypos) = $self->UTM_til_GIF($UTM_N, $UTM_E);

	    $pnavn =~ s/^\(|\)$//g;
 	    my $lnavn = $pnavn;

	    $pnavn =~ tr//aoaAOA/;
	    $pnavn =~ s/ho[gy]sk[ou]len?/HS/ig;
	    $pnavn =~ s/([Uu])niversitet(:?et|ene|s)?/$1niv./g;
	    $pnavn =~ s/([Ii])nstitutt(:?et|ene|s)?/$1nst./g;
	    $pnavn =~ s/([Ff])akultet(:?et|ene|s)?/$1ak./g;
	    $pnavn =~ s/([Ff])orskn?ings?/$1orskn./g;
	    $pnavn =~ s/([Pp])edagog(:?isk)?/$1ed./g;
	    $pnavn =~ s/([Vv])itenskap(:?elige?)?/$1it./g;
	    $pnavn =~ s/([Ii])nternasjonale?/$1nt./g;
	    $pnavn =~ s/([Bb])ibliotek(:?et|ene|s)/$1ibl./g;

	    $self->{'punktdata'}->{$navn}->{'stedtype'} = $stedtype;
	    $self->{'punktdata'}->{$navn}->{'sone'} = $sone;
	    $self->{'punktdata'}->{$navn}->{'xpos'} = $xpos;
	    $self->{'punktdata'}->{$navn}->{'ypos'} = $ypos;
	    $self->{'punktdata'}->{$navn}->{'pnavn'} = $pnavn;
	    $self->{'punktdata'}->{$navn}->{'lnavn'} = $lnavn;
	    $self->{'punktdata'}->{$navn}->{'rutex'} =
	      $rutex = int($xpos / $self->{'rutesx'});
	    $self->{'punktdata'}->{$navn}->{'rutey'} =
	      $rutey = int($ypos / $self->{'rutesy'});

	    # merke oss hvilke institusjoner som er samlet
	    push(@{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}}, $navn);
	    $self->{'rute'}->{$rutex}->{$rutey}->{'textbredde'}=length($pnavn)
	      unless
		defined $self->{'rute'}->{$rutex}->{$rutey}->{'textbredde'} &&
		  length($pnavn) <= $self->{'rute'}->{$rutex}->{$rutey}->{'textbredde'};
	    $self->{'punkt'}->{$navn} = 1;
	}
	elsif ($type eq "SKY") {
	    my($type, $skytype, $navn) = split /\s+/, $linje;
	    $self->{'_skytype'}->{$navn} = $skytype;
	}
	elsif ($type eq "LINJE") {
	    my($type, $linjetype, $navn,
	       $sted1, $sted2,
	       $ptykk, @farge) = split /\s+/,$linje;
	    $self->{'linjedata'}->{$navn}->{'linjetype'} = $linjetype;
	    $self->{'linjedata'}->{$navn}->{'sted1'} = $sted1;
	    $self->{'linjedata'}->{$navn}->{'sted2'} = $sted2;
	    if ($self->{'linjedata'}->{$navn}->{'ptykk'}){
		$self->{'linjedata'}->{$navn}->{'ptykk'} = $ptykk;
		$self->{'linjedata'}->{$navn}->{'farge'} = [@farge];
	    }
	    push(@{$self->{'_linjenavn'}}, $navn);
	}
	else {
	    print STDERR "\Syntaks-feil i linje $. i $self->{'koordfil'}\n";
	}
    }
    close(KOORD);

    if (defined($self->{'hint'})) {
	open (HINT,"$self->{'hint'}") ||
	  die "Kan ikke pne hint $self->{'hint'}: $!\n";

	while (defined($linje = <HINT>)) {
	    next if $linje =~ /^\s*\#|^\s*$/;
	    chomp($linje);
	    $type = $linje;
	    $type =~ s/ .*$// ; # Slett alt utenom frste ord

	    if ($type eq "STED") {
		($type, $stedtype, $navn, $hintx, $hinty, $sone,
		 $UTM_N, $UTM_E, $pnavn) =
		   split /\s+/, $linje, 9;
		if ($sone != $self->{'gsone'}) {
		    my $utmgeo = new utmgeo();
		    ($UTM_N, $UTM_E) =
		      $utmgeo->utmtoutm($UTM_N, $UTM_E, $sone,
					$self->{'gsone'});
		}
		($xpos,$ypos) = $self->UTM_til_GIF($UTM_N, $UTM_E);
		my $rutex = int( $xpos / $self->{'rutesx'} );
		my $rutey = int( $ypos / $self->{'rutesy'} );

		if ($hintx =~ /^([+-]?)(\d+)$/) {
		    $hintx = $1 eq "-" ? $self->{'ant_x'} - $2 : $2;
		} elsif ($hintx =~ /^(\d+)\/(\d+)$/) {
		    $hintx = int(($self->{'ant_x'} * $1)/$2);
		} elsif ($hintx eq 'east') {
		    $hintx = 0;
		} elsif ($hintx eq 'west') {
		    $hintx = $self->{'ant_x'};
		} elsif ($hintx eq 'mid') {
		    $hintx = int($self->{'ant_x'}/2);
		} elsif ($hintx eq 'xpos') {
		    $hintx = $rutex;
		} else {
		    print STDERR "*** Advarsel: ugyldig x-retningshint ".
		      "for $navn ($hintx)\n";
		    $hintx = undef;
		}

		if ($hinty =~ /^([+-]?)(\d+)$/) {
		    $hinty = $1 eq "-" ? $self->{'ant_y'} - $2 : $2;
		} elsif ($hinty =~ /^(\d+)\/(\d+)$/) {
		    $hinty = int(($self->{'ant_y'} * $1)/$2);
		} elsif ($hinty eq 'north') {
		    $hinty = 0;
		} elsif ($hinty eq 'south') {
		    $hinty = $self->{'ant_y'};
		} elsif ($hinty eq 'mid') {
		    $hinty = int($self->{'ant_y'}/2);
		} elsif ($hinty eq 'ypos') {
		    $hinty = $rutey;
		} else {
		    print STDERR "*** Advarsel: ugyldig y-retningshint ".
		      "for $navn ($hinty)\n";
		    $hinty = undef;
		}

		$self->{'steddata'}->{$navn}->{'rutex'} = $rutex;
		$self->{'steddata'}->{$navn}->{'rutey'} = $rutey;
		$self->{'steddata'}->{$navn}->{'hintx'} = $hintx
		  if defined $hintx;
		$self->{'steddata'}->{$navn}->{'hinty'} = $hinty
		  if defined $hinty;
		$self->{'steddata'}->{$navn}->{'pnavn'} = $pnavn;
		$self->{'rute'}->{$rutex}->{$rutey}->{'hintx'}
		  = $hintx if defined $hintx;
		$self->{'rute'}->{$rutex}->{$rutey}->{'hinty'}
		  = $hinty if defined $hinty;
	  }
       }
    }

    if (defined($self->{'last'})) {
	open (LAST,"$self->{'last'}") ||
	  die "Kan ikke pne last $self->{'last'}: $!\n";

	while(defined($linje = <LAST>)) {
	    if ($linje !~ /\s*[\w\d_\.-]+\s+\d+(\s+[\d\.]+){6}/){
		print STDERR "Feil i lastdata :$linje";
		next;
	    }
	    chomp($linje);

	    ($navn, $ptykk, @farge) = split /\s+/,$linje;
	    if (defined($self->{'linjedata'}->{$navn})) {
		$self->{'linjedata'}->{$navn}->{'linjetype'} = "last";
		$self->{'linjedata'}->{$navn}->{'ptykk'} = $ptykk;
		$self->{'linjedata'}->{$navn}->{'farge'} = [@farge];
	    }
	}
	close(LAST);
    }
    return $self;
} # slutt les_Koordinater




# -----------------------------------------------------------
#
# Oversetter UTM-koordinater til GIF-koordinater
# (piksel-koordinater)
#
# -----------------------------------------------------------
#
# ($GIF_x, $GIF_y) = $object->UTM_til_GIF($UTM_nord, $UTM_oest);
#
sub UTM_til_GIF {
      my($self) = shift;
    my($UTM_nord, $UTM_oest) = @_;

    my($GIF_x, $GIF_y);

    $self->{'UTM_areal_hoyde'} =
	$self->{'UTM_areal_nord2'} - $self->{'UTM_areal_nord1'};
    $GIF_x =
	($UTM_oest - $self->{'UTM_areal_oest1'}) /
	    $self->{'skalering'} * $self->{'giftrans'};
    $GIF_y =
	($self->{'UTM_areal_hoyde'} - ($UTM_nord - $self->{'UTM_areal_nord1'}))
	    / $self->{'skalering'} * $self->{'giftrans'};

    $GIF_x = int($GIF_x);
    $GIF_y = int($GIF_y);

    return($GIF_x, $GIF_y);
}

# -----------------------------------------------------------
#
# Setter farge til kartlinjer ved hjelp av RGB verdier
#
# -----------------------------------------------------------
#
# $GDfargeobj = $object->sett_Farge($red_prC, $green_prC, $blue_prC);
#
sub sett_Farge {
    my($self) = shift;
    my($R,$G,$B) = @_;

    my($farge);
    my($nyR,$nyG,$nyB);

    # Gjr prosentverdier til heltallsverdier

    $nyR = $R * 255;
    $nyG = $G * 255;
    $nyB = $B * 255;

    $farge =  $self->{'kartgif'}->colorAllocate($nyR,$nyG,$nyB);

    return ($farge);
}

# -----------------------------------------------------------
#
# Tegner alle punkter med tilhrende stedsnavn p kartet
#
# -----------------------------------------------------------
#
# $object = $object->tegn_Punkt();
#
sub tegn_Punkt {
    my($self) = shift;

    my($xpos, $ypos, $rutex, $rutey, $rute);
    my $ant_x = $self->{'ant_x'};
    my $ant_y = $self->{'ant_y'};
    my($rutesx, $rutesy);
    my($sted, $stedtype, $pnavn);
    my($iconb, $iconh);

    # Hvor mange institusjoner m det vre p et sted fr man lager en
    # tabell?

    my $klumpgrense = $self->{'opt_klump'} || 0;



    $rutesx = $self->{'rutesx'};
    $rutesy = $self->{'rutesy'};

    # Finner institusjoner i alle punktene. (finnes det bedre mter 
    # gjre dette p?) Sortere etter strrelse frst, slik at man kan
    # "samle" se store klumpene frst.

    my @institusjoner = sort
      {
	  my $arutex = $self->{'punktdata'}->{$a}->{'rutex'};
	  my $arutey = $self->{'punktdata'}->{$a}->{'rutey'};
	  my $brutex = $self->{'punktdata'}->{$b}->{'rutex'};
	  my $brutey = $self->{'punktdata'}->{$b}->{'rutey'};
	  # Sortere frst etter "strrelse" p punktet
	  scalar(@{$self->{'rute'}->{$brutex}->{$brutey}->{'klump'}}) <=>
	    scalar(@{$self->{'rute'}->{$arutex}->{$arutey}->{'klump'}})
	      # og deretter alfabetisk
	      || $a cmp $b;
      } keys %{$self->{'punkt'}};


    # G igjennom alle stedene  bygg opp en liste der alle
    # stedsnavnene er samlet pr. koordinat. Om en bestemt koordinat
    # inneholder flere enn $klumpgrense steder, sjekk ogs nabo-rutene,
    # og plasser eventuelle nyfunnede institusjoner i ny koordinat

    if ($klumpgrense != 0) {
        foreach my $sted (@institusjoner) {
	    my $rutex = $self->{'punktdata'}->{$sted}->{'rutex'};
	    my $rutey = $self->{'punktdata'}->{$sted}->{'rutey'};
	    if (scalar(@{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}})>=$klumpgrense ||
		defined $self->{'rute'}->{$rutex}->{$rutey}->{'hintx'} ||
		defined $self->{'rute'}->{$rutex}->{$rutey}->{'hinty'} ) {
		#$self->{'punktdata'}->{$sted}->{'stedtype'}="gruppe_liten.gif";

		foreach my $x (-1..1) {
		    foreach my $y (-1..1) {
			next if $x == 0 && $y == 0;
			my($nabox, $naboy) = ($rutex+$x, $rutey+$y);
			if (defined($self->{'rute'}->{$nabox}->{$naboy}->{'klump'})
			    && ref($self->{'rute'}->{$nabox}->{$naboy}->{'klump'})
			    eq 'ARRAY') {
			    push(@{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}},
				 @{$self->{'rute'}->{$nabox}->{$naboy}->{'klump'}});

			    # flytte eventuelle hint:

			    if (defined $self->{'rute'}->{$nabox}->{$naboy}->{'hintx'}) {
				if (!defined $self->{'rute'}->{$rutex}->{$rutey}->{'hintx'}) {
				    $self->{'rute'}->{$rutex}->{$rutey}->{'hintx'} =
				      $self->{'rute'}->{$nabox}->{$naboy}->{'hintx'};
				} else {
				    print STDERR "*** Advarsel: klump-hint ".
				      "$self->{'rute'}->{$rutex}->{$rutey}->{'navn'} og ".
					"$self->{'rute'}->{$nabox}->{$naboy}->{'navn'} ".
					  "for nrme hverandre!\n";
				}
				delete $self->{'rute'}->{$nabox}->{$naboy}->{'hintx'};
			    }

			    if (defined $self->{'rute'}->{$nabox}->{$naboy}->{'hinty'}) {
				if (!defined $self->{'rute'}->{$rutex}->{$rutey}->{'hinty'}) {
				    $self->{'rute'}->{$rutex}->{$rutey}->{'hinty'} =
				      $self->{'rute'}->{$nabox}->{$naboy}->{'hinty'};
				} else {
				    print STDERR "*** Advarsel: klump-hint ".
				      "$self->{'rute'}->{$rutex}->{$rutey}->{'navn'} og ".
					"$self->{'rute'}->{$nabox}->{$naboy}->{'navn'} ".
					  "for nrme hverandre!\n";
				}
				delete $self->{'rute'}->{$nabox}->{$naboy}->{'hinty'};
			    }

			    delete $self->{'rute'}->{$nabox}->{$naboy}->{'klump'};
			    delete $self->{'_sone'}->{$nabox}->{$naboy};
			}
		    }
		}

		# Srge for  vise et "klump-ikon" av rett type
		if (scalar @{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}}
		    >= $klumpgrense &&
		    scalar @{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}}
		    != 1 ) {
		    $stedtype = "gruppe_middels.gif";
		} else {
		    $stedtype = "gruppe_liten.gif";
		}

		foreach my $sted (@{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}}) {
		    $self->{'punktdata'}->{$sted}->{'stedtype'} = $stedtype;
		    $self->{'punktdata'}->{$sted}->{'rutex'} = $rutex;
		    $self->{'punktdata'}->{$sted}->{'rutey'} = $rutey;
		}


		## Somewhat Grey Magic ahead
		#my @klump_inst = keys
		#  %{({
		#  map {
		#      ($_ => 1)
		#  } @{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}}
		# })};
		#$self->{'rute'}->{$rutex}->{$rutey}->{'klump'} = [ @klump_inst ];
		## Her srget vi for at alle elementene i @klump_inst var unike
	    }
	}
    }

    # Tegne inn rutenett

    $self->_tegn_rutenett() if $self->{'opt_debug'} >= 3;

    # Tegn p alle punkter, dvs. sted, stedsky eller gif-icon

    # i PUNKT-delen skal vi tegne et symbol p kartet for hvert punkt
    # som skal vre med (ikke noe text enda, alts)


  PUNKT:
    foreach $sted (@institusjoner) {

	$stedtype = $self->{'punktdata'}->{$sted}->{'stedtype'};
	$rutex = $self->{'punktdata'}->{$sted}->{'rutex'};
	$rutey = $self->{'punktdata'}->{$sted}->{'rutey'};

	my($xpos, $ypos) = (0 ,0);

	my @punkter_her = @{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}};
	my $ant_pkt_her = scalar(@punkter_her);

	# Prve  finne "gjennomsnittlig posisjon" av alle
	# institusjoner i ($rutex, $rutey).
	foreach my $punkt (@punkter_her) {
	    $xpos += $self->{'punktdata'}->{$punkt}->{'xpos'};
	    $ypos += $self->{'punktdata'}->{$punkt}->{'ypos'};
	    print STDERR "$punkt(xpos,ypos)=($self->{'punktdata'}->{$punkt}->{'xpos'},$self->{'punktdata'}->{$punkt}->{'ypos'}) s=($xpos,$ypos) \n"
	      if $self->{'opt_debug'} >= 2;
	}

	$xpos = int($xpos/$ant_pkt_her);
	$ypos = int($ypos/$ant_pkt_her);

	foreach my $punkt (@punkter_her) { # justere {x,y}pos for alle i klumpen
	    $self->{'punktdata'}->{$punkt}->{'xpos'} = $xpos;
	    $self->{'punktdata'}->{$punkt}->{'ypos'} = $ypos;
	}

	# Hvert punkt okkuperer en rute

	$self->{'_sone'}->{$rutex}->{$rutey} = 1;

	# Tegn punkt som sted, stedsky eller gif-icon

	if ($stedtype =~ /^sted/) {

	    my $pkt_icon = $self->les_Gif_icon("gruppe_liten.gif");
	    ($iconb,$iconh) = $pkt_icon->getBounds();

	    my $iconxpos = int($xpos+($rutesx/2)-($iconb/2));
	    my $iconypos = int($ypos+($rutesy/2)-($iconh/2));

	    $self->{'kartgif'}->setBrush($pkt_icon);
	    $self->{'kartgif'}->setPixel($xpos, $ypos, gdBrushed);

	    # reservere ruter som ikonet overlapper.
	    my $hxrute = int( ($iconxpos + ($iconb/2) )/$rutesx );
	    my $hyrute = int( ($iconypos + ($iconh/2) )/$rutesy );
	    my $lxrute = int( ($iconxpos - ($iconb/2) )/$rutesx );
	    my $lyrute = int( ($iconypos - ($iconh/2) )/$rutesy );

	    $self->reserver_ruter($lxrute, $lyrute, $hxrute, $hyrute);


	    ## Okkuper alle rutene punktet overlapper
	    #my $rutex1 = int((($xpos-($self->{'pkt_storrelse'}/2))/$rutesx));
	    #my $rutey1 = int((($ypos-($self->{'pkt_storrelse'}/2))/$rutesy));
	    #my $rutex2 = int((($xpos+($self->{'pkt_storrelse'}/2))/$rutesx));
	    #my $rutey2 = int((($ypos+($self->{'pkt_storrelse'}/2))/$rutesy));
	    #$self->reserver_ruter($rutex1, $rutey1,
		#		  $rutex2, $rutey2);

	    #$self->{'kartgif'}->arc($xpos, $ypos,
		#		    $self->{'pkt_storrelse'},
		#		    $self->{'pkt_storrelse'},
		#		    0, 360,
		#		    $self->{'punktfarger'}->{'svart'});
	    #$self->{'kartgif'}->fillToBorder($xpos, $ypos,
		#			     $self->{'punktfarger'}->{'svart'},
		#			     $self->{'punktfarger'}->{'rod'});

	    
	} elsif ($stedtype eq "sky") { # 
#	    $self->{'kartgif'}->arc($xpos, $ypos,
#				    $self->{'pkt_storrelse'},
#				    $self->{'pkt_storrelse'},
#				    0, 360,
#				    $self->{'punktfarger'}->{'svart'});
#	    $self->{'kartgif'}->fillToBorder($xpos, $ypos,
#					     $self->{'punktfarger'}->{'svart'},
#					     $self->{'punktfarger'}->{'rod'});
	    $self->{'kartgif'}->arc($xpos, $ypos,
				    $self->{'sky_storrelse'},
				    $self->{'sky_storrelse'},
				    0, 360, $self->{'punktfarger'}->{'rod'});
	} elsif ($stedtype eq "satelitt") {
	    $self->{'kartgif'}->arc($xpos, $ypos,
				    $self->{'sky_storrelse'}*1.7,
				    $self->{'sky_storrelse'}*1.3,
				    0, 360,
				    $self->{'punktfarger'}->{'blaa'});
	} elsif ($stedtype eq "punkt") {
	    # tegn ingenting
	} elsif (!$self->{'_skytype'}->{$sted}) {

	    my $pkt_icon = $self->les_Gif_icon($stedtype);
	    ($iconb,$iconh) = $pkt_icon->getBounds();

	    my $iconxpos = int($xpos+($rutesx/2)-($iconb/2));
	    my $iconypos = int($ypos+($rutesy/2)-($iconh/2));

	    $self->{'kartgif'}->setBrush($pkt_icon);
	    $self->{'kartgif'}->setPixel($xpos, $ypos, gdBrushed);

	    # reservere ruter som ikonet overlapper.
	    my $hxrute = int( ($iconxpos + ($iconb/2) )/$rutesx );
	    my $hyrute = int( ($iconypos + ($iconh/2) )/$rutesy );
	    my $lxrute = int( ($iconxpos - ($iconb/2) )/$rutesx );
	    my $lyrute = int( ($iconypos - ($iconh/2) )/$rutesy );

	    $self->reserver_ruter($lxrute, $lyrute, $hxrute, $hyrute);
	}

	if ($self->{'_skytype'}->{$sted}) {
	    if ($self->{'_skytype'}->{$sted} =~ /satelitt/) {
		$self->{'kartgif'}->arc($xpos, $ypos,
					$self->{'sky_storrelse'}*1.7,
					$self->{'sky_storrelse'}*1.3,
					0, 360,
					$self->{'punktfarger'}->{'blaa'});
	    } else {
		# tegner 4 halvsirkler om et kvadrat om punktet
		my $r = $self->{'sky_storrelse'};
		my $dr = $r/2;
		$self->{'kartgif'}->arc($xpos-$dr, $ypos, $r, $r, 90, 270,
					$self->{'punktfarger'}->{'rod'});
		$self->{'kartgif'}->arc($xpos, $ypos+$dr, $r, $r, 0, 180,
					$self->{'punktfarger'}->{'rod'});
		$self->{'kartgif'}->arc($xpos+$dr, $ypos, $r, $r, 270, +90,
					$self->{'punktfarger'}->{'rod'});
		$self->{'kartgif'}->arc($xpos, $ypos-$dr, $r, $r, 180, +0,
					$self->{'punktfarger'}->{'rod'});
	    }
	}
	print STDERR "Punkt ferdig : $sted\n" if $self->{'opt_debug'} >= 1;

    }

    # Slutt p PUNKT


    # sett tekst p punkter

    return if $self->{'opt_n'}; # ikke tegn p texter

    # Okkuper alle rutene nederst pga tekster der
    $self->reserver_ruter(0, $self->{'ant_y'}-1,
			  $self->{'ant_x'}, $self->{'ant_y'});


    # Okkuper alle rutene lengst til hyre for  unng tekst "utenfor"
    # bildet
    $self->reserver_ruter($self->{'ant_x'}, 0,
			  $self->{'ant_x'}, $self->{'ant_y'});


    # Reservere et "fristed" et par linjer nederst til hyre, s
    # mlestokken p kartet ikke blir skrevet over
    my $mstokk_reservering = 15;
    $self->reserver_ruter($self->{'ant_x'}-$mstokk_reservering,
			  $self->{'ant_y'}-3,
			  $self->{'ant_x'}, $self->{'ant_y'});

    # Okkuper alle rutene lengst til venstre for  unng tekst for
    # nrme kanten
    $self->reserver_ruter(0, 0,
			  0, $self->{'ant_y'});



    # Okkuper alle rutene verst for  unng tekst for nrme kanten
    $self->reserver_ruter(0, 0,
			  $self->{'ant_x'}, 0);



    if (defined($self->{'textpos'})) { # S vi kan ta vare p tekst-koord.
	open(TEXTPOS, ">$self->{'textpos'}") || die "Fikk ikke pnet $self->{'textpos'}: $!\n";
    }


    ### Skriv stedsnavn til alle punktene p kartet

    if ($klumpgrense != 0) {

	# Frst alle stedene som ikke skal plasseres i boks, deretter alle
	# stedene i boks

	my(@hintklumper, @klumper, @smaaklumper, @steder);
	foreach (sort @institusjoner) {
	    my $rutex = $self->{'punktdata'}->{$_}->{'rutex'};
	    my $rutey = $self->{'punktdata'}->{$_}->{'rutey'};
	    if (defined $self->{'rute'}->{$rutex}->{$rutey}->{'hintx'} ||
		defined $self->{'rute'}->{$rutex}->{$rutey}->{'hinty'}) {
		push(@hintklumper, $_);
	    } elsif (scalar(@{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}})
		     > $klumpgrense ) {
		push(@klumper, $_);
	    } elsif (scalar(@{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}})
		     <= $klumpgrense ) {
		push(@smaaklumper, $_);
	    }
	}

	####
	#### "Klumper" - "steder med mange institusjoner"
	####

	# Fr vi begynner med klumpene, m vi ha skikkelig oversikt over
	# hvilke steder p kartet som er ledige.

	# Her klumpgger vi opp et "ledig-plass-kart" som skal fortelle
	# oss hvor mange ruter som er ledig til hyre for og nedenfor
	# et gitt punkt p kartet.
	#
	# Et lite kart kan se ut som flgende:
	#
	# 4221XY31
	# 34132221
	# 21XY1211
	# 1211XYXY

	for (my $ry = $ant_y; $ry >= 0; $ry--) {
	    for (my $rx = $ant_x; $rx >= 0; $rx--) {
		if (defined $self->{'_sone'}->{$rx}->{$ry}) {
		    $rutekart->[$rx]->[$ry]->[0] = 0;
		    $rutekart->[$rx]->[$ry]->[1] = 0;
		} else {
		    if (defined($rutekart->[$rx+1]->[$ry]->[0])) {
			$rutekart->[$rx]->[$ry]->[0] =
			  $rutekart->[$rx+1]->[$ry]->[0] + 1;
		    } else {
			$rutekart->[$rx]->[$ry]->[0] = 1;
		    }
		    if (defined($rutekart->[$rx]->[$ry+1]->[1])) {
			$rutekart->[$rx]->[$ry]->[1] =
			  $rutekart->[$rx]->[$ry+1]->[1] + 1
		      } else {
			  $rutekart->[$rx]->[$ry]->[1] = 1;
		      }
		}
	    }
	}

	my @sortertesmaaklumper = sort
	  {
	      my $arutex = $self->{'punktdata'}->{$a}->{'rutex'};
	      my $arutey = $self->{'punktdata'}->{$a}->{'rutey'};
	      my $brutex = $self->{'punktdata'}->{$b}->{'rutex'};
	      my $brutey = $self->{'punktdata'}->{$b}->{'rutey'};
	      # Sortere frst etter hvor mange institusjoenr
	      # pr. punkt, i stigende rekkeflge, s etter bredde p
	      # teksten, i stigende rekkeflge.
	      scalar(@{$self->{'rute'}->{$brutex}->{$brutey}->{'klump'}}) <=>
		scalar(@{$self->{'rute'}->{$arutex}->{$arutey}->{'klump'}}) ||
		  length($self->{'rute'}->{$arutex}->{$arutey}->{'textbredde'}) <=>
		    length($self->{'rute'}->{$brutex}->{$brutey}->{'textbredde'});

	  } @smaaklumper;

	# s sorterer vi klumpene s de strste kommer frst

	my @sorterteklumper = sort
	  {
	      my $arutex = $self->{'punktdata'}->{$a}->{'rutex'};
	      my $arutey = $self->{'punktdata'}->{$a}->{'rutey'};
	      my $brutex = $self->{'punktdata'}->{$b}->{'rutex'};
	      my $brutey = $self->{'punktdata'}->{$b}->{'rutey'};
	      # Sortere frst etter hvor mange institusjoenr
	      # pr. punkt, i synkende rekkeflge.
	      scalar(@{$self->{'rute'}->{$arutex}->{$arutey}->{'klump'}}) <=>
		scalar(@{$self->{'rute'}->{$brutex}->{$brutey}->{'klump'}});

	  } @klumper;

	# til slutt sorterer vi klumpene som har ftt plasseringshint

	my @sortertehintklumper = sort
	  {
	      my $arutex = $self->{'punktdata'}->{$a}->{'rutex'};
	      my $arutey = $self->{'punktdata'}->{$a}->{'rutey'};
	      my $brutex = $self->{'punktdata'}->{$b}->{'rutex'};
	      my $brutey = $self->{'punktdata'}->{$b}->{'rutey'};
	      # Sortere frst etter hvor mange institusjoenr
	      # pr. punkt, i synkende rekkeflge.
	      scalar(@{$self->{'rute'}->{$brutex}->{$brutey}->{'klump'}}) <=>
		scalar(@{$self->{'rute'}->{$arutex}->{$arutey}->{'klump'}});

	  } @hintklumper;

	my $been_there = {};	    # for  holde styr p hvor vi har
                                    # vrt

      KLUMP:
	foreach my $sted (@sortertehintklumper, @sortertesmaaklumper, @sorterteklumper) {
	    my $rutex = $self->{'punktdata'}->{$sted}->{'rutex'};
	    my $rutey = $self->{'punktdata'}->{$sted}->{'rutey'};

	    next KLUMP if $been_there->{$rutex}->{$rutey};
	    $been_there->{$rutex}->{$rutey} = 1;

	    $self->_plasser_klump($sted, $rutekart, 31);
	}

    } else {

	$self->_plasser_sted(\@institusjoner);

    }

  EOKLUMPER:

    if (defined($self->{'textpos'})) {
	close(TEXTPOS)
	  || warn "Fikk ikke lukket $self->{'textpos'}!?!: $!\n";
    }

    # legg ut kordinater for textene til kartmap

    #goto EOGRID;

    if ($klumpgrense > 0 && $self->{'opt_debug'} >= 3) {
	print STDERR "\n  .",(".."x50),"\n  :",(" 0 1 2 3 4 5 6 7 8 9"x5),"\n";

	foreach my $ry (0 .. $ant_y) {
	    printf STDERR "%2s:", $ry;
	    foreach my $rx (0 .. $ant_x) {
		my $cx = $rutekart->[$rx]->[$ry]->[0];
		my $cy = $rutekart->[$rx]->[$ry]->[1];
		my $px = $cx;
		my $py = $cy;
		$px = chr($cx+87) if $cx >= 10 && $cx <= 35;
		$py = chr($cy+87) if $cy >= 10 && $cy <= 35;
		$px = chr($cx+29) if $cx >= 36;
		$py = chr($cy+29) if $cy >= 36;
		$px = ' ' unless $px;
		$py = ' ' unless $py;
		printf STDERR "%s%s", $px, $py;
	    }
	    print STDERR "\n";
	}
    }

  EOGRID:

    return $self;
}				# slutt tegn_Punkt


sub _plasser_klump {
    my $self = shift;
    my $sted = shift;
    my $rutekart = shift;
    my $tbreddeforslag = shift; # hvor mange bokstaver bred skal tekstboksen vre?

    my $rutex = my $rutepx = $self->{'punktdata'}->{$sted}->{'rutex'};
    my $rutey = my $rutepy = $self->{'punktdata'}->{$sted}->{'rutey'};

    my $textref = $self->{'rute'}->{$rutex}->{$rutey}->{'klump'};

    $self->{'rute'}->{$rutex}->{$rutey}->{'textbredde'} =
      $self->_masser_text($self->{'rute'}->{$rutex}->{$rutey}->{'klump'},
			   $tbreddeforslag);

    my($antxr, $antyr) = $self->_finn_blokkxy($sted);

    $rutepx = $self->{'rute'}->{$rutex}->{$rutey}->{'hintx'}
      if defined $self->{'rute'}->{$rutex}->{$rutey}->{'hintx'};

    $rutepy = $self->{'rute'}->{$rutex}->{$rutey}->{'hinty'}
      if defined $self->{'rute'}->{$rutex}->{$rutey}->{'hinty'};


    ($rutepx, $rutepy) = # Plassering av textboks p kartet
      $self->_finn_ledig_boks($rutekart, $rutepx, $rutepy, $antxr, $antyr);


    if ($self->_ledig_plass_p($rutekart, $rutepx, $rutepy,
			      $antxr, $antyr)) {

	$self->_oppdater_rutekart($rutekart,
				  $rutepx, $rutepy,
				  $antxr, $antyr);

	$self->_skriv_textblokk($self->{'rute'}->{$rutex}->{$rutey}->{'klump'},
				$rutepx, $rutepy,
				$self->{'punktdata'}->{$sted}->{'xpos'},
				$self->{'punktdata'}->{$sted}->{'ypos'},
				$antxr, $antyr);
    } else {
	print STDERR
	  "*** Advarsel: Textboks strrelse ($antxr:$antyr) [@{$textref}]"
	    ."i ($rutex, $rutey) kan ikke plasseres i kartet!\n";
    }
}


sub reserver_ruter {
    my $self = shift;
    my($rutex1, $rutey1, $rutex2, $rutey2) = @_;

    my $rutesx = $self->{'rutesx'};
    my $rutesy = $self->{'rutesy'};

    printf STDERR "reserver_ruter(%d, %d, %d, %d)\n => (%d,%d) (%d,%d)\n",
	$rutex1, $rutey1, $rutex2, $rutey2,
	  $rutex1*$rutesx,
	    $rutey1*$rutesy,
	      $rutex2*$rutesx+$rutesx-1,
		$rutey2*$rutesy+$rutesx-1
		  if $self->{'opt_debug'} >= 2;

    $self->{'kartgif'}->rectangle($rutex1*$rutesx,
				  $rutey1*$rutesy,
				  $rutex2*$rutesx+$rutesx-1,
				  $rutey2*$rutesy+$rutesx-1,
				  $self->{'punktfarger'}->{'gronn'})
      if $self->{'opt_debug'} >= 3;


    foreach my $xrute ($rutex1 .. $rutex2) {
	foreach my $yrute ($rutey1 .. $rutey2) {
	    $self->{'_sone'}->{$xrute}->{$yrute} = 1;
	}
    }
    return $self;
}


# ########################################################################
#
# plasserer ut klumper p orginalmten.
#
# $object = $object->_plasser_sted(\@steder);
#
sub _plasser_sted {
    my $self = shift;
    my $steder = shift; # m vre en referanse til array
    my $rutesx = $self->{'rutesx'};
    my $rutesy = $self->{'rutesy'};
    my($text_retning, $tegnlinje);
    my($EE, $SE, $SS, $SW, $WW, $NW, $NN, $NE);

    my $ant_x = $self->{'ant_x'};
    my $ant_y = $self->{'ant_y'};

  STED:
    foreach $sted (sort @{$steder}) {

	my $pnavn = $self->{'punktdata'}->{$sted}->{'pnavn'};
	my $lnavn = $self->{'punktdata'}->{$sted}->{'lnavn'};

	$xpos = $self->{'punktdata'}->{$sted}->{'xpos'};
	$ypos = $self->{'punktdata'}->{$sted}->{'ypos'};
	# ($xpos, $ypos) = split(/,/, $ko{$sted});

	$text_rutex = int ($xpos / $rutesx);
	$text_rutey = int ($ypos / $rutesy);
	$EE = $SW = $WW = $NE = $text_rutex;
	$SE = $SS = $NW = $NN = $text_rutey;
	$text_retning = "EAST";

	$text_rutex++;		# for  f texten i ruta tilhyre for punktet?

	my $lengde = ($self->{'tb'} * length($pnavn));
	my $ant_x_ruter = int(($lengde / $rutesx)+0.5);
	my $ant_y_ruter = 1;

	undef($tegnlinje);

      NYRUTE:

	if (($ant_x_ruter+$text_rutex) >= ($ant_x-1)) {
	    $text_rutex = ($ant_x-1)-$ant_x_ruter;
	}
	if ($text_rutey >= ($ant_y-1)) {
	    $text_rutey = ($ant_y-1);
	}
	if ($text_rutex < 0 && $text_rutey < 0) { # forsk p lkkekontroll
	    print STDERR "Kan ikke plassere tekst for $pnavn i $text_rutex,$text_rutey p %s\n", $self->{'koordfil'};
	    next STED;
	}
	if ($text_rutex < 0) {
	    $text_rutex = 0;
	}
	if ($text_rutey < 0) {
	    $text_rutey = 0;
	}

	for ($rute=$text_rutex;
	     $rute <= ($text_rutex+$ant_x_ruter);
	     $rute++) {
	    if (defined($self->{'_sone'}->{$rute}->{$text_rutey})) {
	      SWITCH: {

		    if ($text_retning eq "EAST") {
			$EE++;
			$SE++;
			$text_rutex = $EE;
			$text_rutey = $SE;
			$text_retning = "SOUTH";
			last SWITCH;
		    }
		    if ($text_retning eq "SOUTH") {
			$SS++;
			$SW--;
			$text_rutex = $SW;
			$text_rutey = $SS;
			$text_retning = "WEST";
			last SWITCH;
		    }
		    if ($text_retning eq "WEST") {
			$WW--;
			$NW--;
			$text_rutex = $WW;
			$text_rutey = $NW;
			$text_retning = "NORTH";
			last SWITCH;
		    }
		    if ($text_retning eq "NORTH") {
			$NN--;
			$NE++;
			$text_rutex = $NE;
			$text_rutey = $NN;
			$text_retning = "EAST";
			last SWITCH;
		    }
		}
		$tegnlinje = 1;
		goto NYRUTE;
	    }
	}

	# Okkuper alle rutene stedsnavnet dekker
	$self->reserver_ruter($text_rutex, $text_rutey,
			      $text_rutex+$ant_x_ruter, $text_rutey);

	my $nyposx = $text_rutex * $rutesx;
	my $nyposy = $text_rutey * $rutesy;

	if (defined($self->{'textpos'})) {
	    printf TEXTPOS "PUNKT text %-15s gif %3d %3d (%s)\n",
	      $sted, $nyposy, $nyposx, $lnavn;
	}

	# Tegn linje fra punkt til stedsnavn dersom stedsnavn ikke
	# ligger ved siden av punktet

	if (defined($tegnlinje)) {
	    $self->{'kartgif'}->line($xpos,
				     $ypos,
				     $nyposx,
				     $nyposy+int($self->{'th'}/2),
				     $self->{'punktfarger'}->{'svart'});
	}
	$self->{'kartgif'}->string(gdSmallFont,
				   $nyposx+2, $nyposy,
				   $pnavn,
				   $self->{'punktfarger'}->{'svart'});
    }

    return $self;
}





# -----------------------------------------------------------
#
# finner frste x og y-koordinat der en "boks" kan plasseres
# - f.eks skal jeg plassere en 6x7 boks, og denne funksjonenen
# skal da finne punktet X der hele boksen fr plass til hyre
# for og nedenfor punktet.
#
#  X--6-+
#  |    |
#  |    |
#  7    |
#  |    |
#  |    |
#  +----+
#
# Algoritmeforslag:
#   0.0.0 MERK! Vi antar ar alle "smsteder" er blitt utplassert, og
#         det KUN er "klumper" igjen!
#   1.0.0 Finn det hjrnet som er lengst unna i x, men nrmest i y,
#         plasser "markr" her (markren er det punktet vi leter fra
#         for  sjekke om vi har plass til boksen).
#   1.1.0 Finn det hjrnet som er nrmest i x og y-retning, og
#         plasser "markr" her (markren er det punktet vi leter fra
#         for  sjekke om vi har plass til boksen).
#   1.2.0 Kvadrantinndeling frst! hver storklump, sortert etter
#         strrelse (posisjon?) fr tildelt en kavadrant der boksen
#         helst skal plasseres. markr starter sk i det mest fjerne
#         punktet i kvadranten.
#   1.3.0 Vi starter lengst sydst, og lager to lister for hver rute-
#         rad, der vi velger annenhver klump (st/vest) og plasserer dem
#         henholdvis i i "-listen" eller "v-listen". Dette gjr vi
#         for hver rad, til alle klumper har havnet i en liste hver.
#         deretter sorterer vi listene slik at de "tyngste" listene
#         kommer frst, og plasserer markren i ekstremiteten til
#         den retningen klumpen er i ("-liste" eller "v-liste").
#   1.3.1 ... sorterer slik at de listene med "tyngst klumper" kommer
#         frst ...
#   2.0.0 Juster markrposisjoenen i x og/eller y-retning om vi er i
#         2., 3. eller 4. kvadrant, slik at vi fr plass til boksen.
#   3.0.0 Sjekk om boksen fr plass her, hvis ja, marker opptatt og
#         avslutt funksjonen! :D
#         a) nr du sjekker: se p vekt-tabellen for rutene for 
#            finne ut om det er ledig plass!
#         b) hvis det ikke er plass, flytt markren i en spiralformet
#            bane runt utgangspunktet (noe lignende algoritmin som
#            brukes for  finne ledig plass tor enkeltsteder) og sjekk
#            for ledig plass.
#   3.0.1 b) hvis det ikke er plass, sk frst i y-retning om du kan f
#            plass. hvis ja, s er alt fint, ellers juster et par
#            punkter i x-retning for  se om det er ledig plass der.
#   4.0.0 Ske-etter-ledig-plass-algoritme:

#   4.0.1 Start i (0,0) eller (maxx,maxy) hjrnet nom er
#         nrmest,

#   4.0.2 Sjekk om det er ledig plass der.
#         Hvis omrdet er ledig, plasser greiene her!
#         Hvis omrde X:Y er helt opptatt,
#         - flytt boksx i x-retning, mot sentrum, og sjekk om det
#         er ledig der der.
#         - eller flytt boksy i y-retning mot sentrum, og sjekk om
#         det er ledig der.
#         - eller flytt boksx i x-retning _fra_ sentrum, og sjekk
#         om det er ledig der.

#         Hvis det finnes ledige plasser innenfor, finn ut
#         hvilke posisjoner som kan brukes. F.eks:

########################################################################
########################################################################
########################################################################
#
#   4.0.3 Starter i et hjrne, og sjekker hvor mange ledige plasser
#         mot hyre og nedover det er. Hvis ledige plasser (x:y) er
#         mindre enn (b:h), flytt nok ruter frst i x-retning mot
#         sentrum, deretter i y-retning mot sentrum, og til slutt i x
#         og y retning s det (muligens) blir plass. Sjekk om disse er
#         ledige, hvis ja, plasser i frste ledige, ellers prv igjen
#         videre...
#
########################################################################
########################################################################
########################################################################


#   4.0.4 Jeg nsker  plassere et 5x6, sjekker et bestemt
#         omrde, og finner ut at flgende felt er opptatt:
#
#                    ,- y
#                   /,- y-1
#                  //,- y-5
#                 ///,- y-3
#                ////,- y-3
#               |||||
#
#         x   -  - x
#         x-3 - # - x+3
#         x   -  - x
#         x-2 - ## - x+5
#         x-2 - ## - x+5
#         x-4 - ## - x+5
#
#               |||||
#         y    -'////
#         y+6  -'///
#         y+2  -'//
#         y+5  -'/
#         y+6  -'
#
#
#         -X   = -4
#         +X   = +5
#         -Y   = -4
#         +Y   = +5
#         -X-Y = -3-1 eller -b-h ?
#         +X-Y = +3-3 eller +b-h ?
#         -X+Y = -4+5 eller -b+h ?
#         +X+Y = +5+5 eller +b+h ?





#   4.0.5 (alternativ, smartere skerekkeflge? - Rekkeflgen av
#         hjrner  ske i er nrmest-lengst unna.)
#
#         X 1 3   3 1 X
#         2 4 6   6 4 2
#         5 7 8   8 7 5
#
#         5 7 8   8 7 5
#         2 4 6   6 4 2
#         X 1 3   3 1 X
#


#   4.0.6 Nok en alternativ skerekkeflge, men denne gangen ut i fra
#         det punktet boksen skal representere.
#
#                  1 3 5
#                  2 X 7
#                  4 6 8
#
#                      sentrum
#                           \
#                            \
#
#         1 => mot (max(x), max(y))
#         2 => mot (max(x), ca0(y))  # br vi g rett mot y, eller m vi ikke?
#         3 => mot (ca0(x), max(y))
#         4 => mot (max(x), min(y))
#

#
#
# -----------------------------------------------------------





# -----------------------------------------------------------
#
# Finner strrelsen i $ruterx:$rutery som kreves for  vise $textref
# arrayet.
#
# $textref m vre en referanse til en array med _ferdigmassert_ text.
#
# -----------------------------------------------------------
#
# ($ant_x, $ant_y) = $object->_finn_blokkxy($textref);
#
sub _finn_blokkxy {
    my $self = shift;
    my $sted = shift;
    my $text_nudge = $self->{'_nudge'};

    my $rutex = $self->{'punktdata'}->{$sted}->{'rutex'};
    my $rutey = $self->{'punktdata'}->{$sted}->{'rutey'};

    print STDERR "_finn_blokkxy($sted) ($self->{'punktdata'}->{$sted}->{'pnavn'})\n" if $self->{'opt_debug'} >= 2;

    my $textbredde = $self->{'rute'}->{$rutex}->{$rutey}->{'textbredde'};

    my $ant_x_ruter =
      int( (($textbredde*$self->{'tb'})+$text_nudge)/$self->{'rutesx'}+0.5 );
    my $ant_y_ruter = scalar
      map { @{$self->{'punktdata'}->{$_}->{'mnavn'}} }
	@{$self->{'rute'}->{$rutex}->{$rutey}->{'klump'}};

    return($ant_x_ruter, $ant_y_ruter)
}





# -----------------------------------------------------------
#
# Rutine for  sjekke om deter ledig plass til en AxB stor boks
# med utgangspunkt (X, Y)
#
# -----------------------------------------------------------
#
# $boolsk=$object->_ledig_plass_p($rutekart, $rutex, $rutey, $boksx, $boksy);
#
sub _ledig_plass_p {
    my $self = shift;
    my($rutekart, $rutex, $rutey, $boksx, $boksy) = @_;

    return 0 unless $rutex >= 0 && $rutex <= $self->{'ant_x'} &&
      $rutey >= 0 && $rutey <= $self->{'ant_y'};
    return 0 unless(defined $rutekart->[$rutex]->[$rutey]->[0] &&
		    defined $rutekart->[$rutex]->[$rutey]->[1]);
    if ($rutekart->[$rutex]->[$rutey]->[0] >= $boksx &&
	$rutekart->[$rutex]->[$rutey]->[1] >= $boksy) {
	my $minxy = $boksx<$boksy?$boksx:$boksy;
	for(my $dxy = 1; $dxy < $minxy; $dxy++) {
	    return 0 unless
	      ($rutekart->[$rutex+$dxy]->[$rutey+$dxy]->[0] >= $boksx-$dxy &&
	       $rutekart->[$rutex+$dxy]->[$rutey+$dxy]->[1] >= $boksy-$dxy);
	}
	return 1;
    } else {
	return 0;
    }
}

# -----------------------------------------------------------
#
# Oppdatere rute- og sone-kartet (dvs, srge for at en ny boks fr
# reservert ruter p kartet og at "ledig plass"-tabellen blir justert.
#
# -----------------------------------------------------------
#
# $object = $object->_oppdater_rutekart($rutekart,$rutex,$rutey,$boksx,$boksy);
#
# $rutex, $rutey er for plassering av boks
# $boksx, $boksy er strrelsen av boksen som skal reserveres.
#   ALLE VERDIER ER I ANTALL RUTER!
#
sub _oppdater_rutekart {
    my $self = shift;
    my($rutekart, $rutex, $rutey, $boksx, $boksy) = @_;
    my $ant_x = $self->{'ant_x'};
    my $ant_y = $self->{'ant_y'};

  Y: for (my $ry = $rutey+$boksy-1; $ry >= 0; $ry--) {
      X: for (my $rx = $rutex+$boksx-1; $rx >= 0; $rx--) {
	    next Y if($rx < $rutex && $ry < $rutey);
	    if ($rx >= $rutex && $ry >= $rutey) { # 1
		# innenfor boksen? reserver ruta!
		print STDERR "*** Advarsel: Rute $rx:$ry allerede reservert!\n"
		  if defined $self->{'_sone'}->{$rx}->{$ry};

		$rutekart->[$rx]->[$ry]->[0] = 0;
		$rutekart->[$rx]->[$ry]->[1] = 0;
		$self->{'_sone'}->{$rx}->{$ry} = 1;
	    } elsif ($rx < $rutex) { # 2
		if (defined $self->{'_sone'}->{$rx}->{$ry}) {
		    next Y if defined $countx[$ry];
		    $countx[$ry]++;
		    next Y ;
		} else {
		    $rutekart->[$rx]->[$ry]->[0] =
		      $rutekart->[$rx+1]->[$ry]->[0] + 1 || 1;
		}
		$countx[$rx]++;
	    } elsif ($ry < $rutey) { # 3
		if (defined $self->{'_sone'}->{$rx}->{$ry} || $rx < $rutex) {
		    next Y if defined $county[$rx];
		    $county[$rx]++;
		} else {
		    $rutekart->[$rx]->[$ry]->[1] =
		      $rutekart->[$rx]->[$ry+1]->[1] + 1 || 1;
		}
	    }
	}
    }
    return $self;
}



##
## Gammel (ineffektiv) versjon
##
#sub _oppdater_rutekart {
#    my $self = shift;
#    my($rutekart, $rutex, $rutey, $boksx, $boksy) = @_;
#    my $ant_x = $self->{'ant_x'};
#    my $ant_y = $self->{'ant_y'};
#    my($ry, $rx);
#
#    for ($ry = $ant_y; $ry >= 0; $ry--) {
#	for ($rx = $ant_x; $rx >= 0; $rx--) {
#	    if ($rx >= $rutex && $rx <= $rutex+$boksx &&
#		$ry >= $rutey && $ry <= $rutey+$boksy) {
#		$self->{'_sone'}->{$rx}->{$ry} = 1;
#		$rutekart->[$rx]->[$ry]->[0] = 0;
#		$rutekart->[$rx]->[$ry]->[1] = 0;
#	    } elsif (defined($self->{$rx}->{$ry})) {
#		$rutekart->[$rx]->[$ry]->[0] = 0;
#		$rutekart->[$rx]->[$ry]->[1] = 0;
#	    } else {
#		if (defined($rutekart->[$rx+1]->[$ry]->[0])) {
#		    $rutekart->[$rx]->[$ry]->[0] =
#		      $rutekart->[$rx+1]->[$ry]->[0] + 1;
#		} else {
#		    $rutekart->[$rx]->[$ry]->[0] = 1;
#		}
#		if (defined($rutekart->[$rx]->[$ry+1]->[1])) {
#		    $rutekart->[$rx]->[$ry]->[1] =
#		      $rutekart->[$rx]->[$ry+1]->[1] + 1;
#		} else {
#		    $rutekart->[$rx]->[$ry]->[1] = 1;
#		}
#	    }
#	}
#    }
#    return $self;
#}


# -----------------------------------------------------------
#
# Funksjon som skal g igjennom en text-liste, og srge for at den er
# s smal som mulig (tar minst mulig areal), ved  la lange text-
# strenger g over flere linjer.
#
# -----------------------------------------------------------
#
# $textbredde = $object->_massr_text($textref, $breddeforslag);
#
sub _masser_text {
    my $self = shift;
    my $textref = shift;
    my $tbreddeforslag = shift || 0;
    my $prefix = "";
    my $nytextbredde = 0;

    # Frst nsker vi  finne ut lengste enkelt-ord for  bestemme
    # hvor bred (i antall bokstaver) texten kan vre

    # lysegr magi: sorter etter lengste sub-streng i pnavn'ene til
    # navn for  finne minste textbredde som trengs for  vise
    # texten
    my $textbredde = # 1 +
      ( sort { $a <=> $b }
	map { length }
	split(/\s+|-/,
	      join(" ",
		   map { $self->{'punktdata'}->{$_}->{'pnavn'} }
		   @{$textref}
		  )
	     )
      )[0];

    $textbredde = $tbreddeforslag if $tbreddeforslag > $textbredde;

    # Ommarangering av text, slik at den passer s godt som mulig inn
    # i bredden vi nsker ha p textboksen

    foreach my $sted (@{$textref}) {
	my $text = $self->{'punktdata'}->{$sted}->{'pnavn'};
	my @nytext = ();

	if(length($text) > $textbredde) {

	    $text =~ s/-\s*/- /g;
	    my @texter	= split(/ /, $text);

	    my $a = shift @texter;
	    while (my $b = shift @texter) {
		if($a =~ /-$/ && length("$a$b") <= $textbredde) {
		    $a = "$a$b";
		} elsif (length("$a $b") <= $textbredde) {
		    $a = "$a $b";
		} else {
		    push(@nytext, $a);
		    $a = "$prefix$b";
		}
	    }
	    push(@nytext, $a);
	} else {
	    push(@nytext, $text);
	}

	$self->{'punktdata'}->{$sted}->{'mnavn'} = [ @nytext ];

	my $bredde = # finne den lengste linja i arrayet
	  ( sort { $b <=> $a }
	    map { length }
	    @nytext
	  )[0];

	$nytextbredde = $bredde if $bredde > $nytextbredde;
    }

    return $nytextbredde ;
}




# -----------------------------------------------------------
#
# Skriver v.hj.a. GD en x:y stor blokk med text p gif-bildet, p
# ruteisjon X:Y.
#
# $textref skal vre en arraysom inneholder _$navn_ (ikke $pnavn)!
# $xpos og $ypos er posisjonen p kartet boksen skal peke p
# $rutex og $rutey er plasseringen p textboksen
# -----------------------------------------------------------
#
# $objekt=$object->_skriv_textblokk($textref, $rutex, $rutey, $xpos, $ypos);
#
sub _skriv_textblokk {
    my $self = shift;
    my($textref, $rutex, $rutey, $xpos, $ypos) = @_;

    my $rutesx = $self->{'rutesx'};
    my $rutesy = $self->{'rutesy'};
    my $linjenum = 0;
    my $hoyre_align = 0;
    my $sted = $textref->[0];
    my $pkt_icon;

    my $text_nudge = $self->{'_nudge'};

    my $klump_rutex = $self->{'punktdata'}->{$sted}->{'rutex'};
    my $klump_rutey = $self->{'punktdata'}->{$sted}->{'rutey'};

    my $textbredde = # bokstaver
      $self->{'rute'}->{$klump_rutex}->{$klump_rutey}->{'textbredde'};
    my $texthoyde = scalar map { @{$self->{'punktdata'}->{$_}->{'mnavn'}} } @{$textref};

    my $boksbredde = $textbredde * $self->{'tb'} + $text_nudge; # pixler
    my $bokshoyde = $texthoyde * $self->{'th'};

    if ($xpos > ($boksbredde/2)+$rutex*$rutesx) {
	$hoyre_align = 1;
    }

    if ($texthoyde == 1) {
	$pkt_icon = $self->{'micro_icon'};
    } else {
	$pkt_icon = $self->{'pkt_icon'};
    }

    $self->{'kartgif'}->setBrush( $pkt_icon );
    my($iconb, $iconh) = $pkt_icon->getBounds();

    my $dot_nudge = $iconb;

    # Evt. tegn inn en lysebl boks under teksten, for  gjre den mer
    # leselig..
    $self->{'kartgif'}->filledRectangle($rutex*$rutesx+1,
					$rutey*$rutesy+1,
					$rutex*$rutesx+$boksbredde-1,
					$rutey*$rutesy+$bokshoyde-1,
					$self->{'punktfarger'}->{'lyseblaa'})
      if $self->{'opt_textbg'};

    foreach my $sted (sort
		      {$self->{'punktdata'}->{$a}->{'lnavn'} cmp
			 $self->{'punktdata'}->{$b}->{'lnavn'}} @{$textref}) {
	my $forstelinje = 1;

	my $linjeant = scalar @{$self->{'punktdata'}->{$sted}->{'mnavn'}};

	if (defined($self->{'textpos'})) { # S vi kan ta vare p koord.
	    printf TEXTPOS "AREAL rektangel %-15s %3d %3d %3d %3d (%s)\n",
	      $sted, $rutex*$rutesx,
		($linjenum+$rutey)*$rutesy,
		  $rutex*$rutesx+$boksbredde,
		    (($linjenum+$rutey+$linjeant)*$rutesy)-1,
		      $self->{'punktdata'}->{$sted}->{'lnavn'};
	}


	# Ommbler teksten om den skal vre hyrejustert, og finn
	# koortdinatene til stedene hvor tekst eller streker skal
	# tegnes.
	foreach my $linje (@{$self->{'punktdata'}->{$sted}->{'mnavn'}}) {
	    $linje =~ tr//aoaAOA/;

	    if ($hoyre_align) {
		my $tmp = " "x($textbredde);
		$linje =~ /^(.*)$/;
		substr($tmp, -length("$1")) = "$1";
		$linje = $tmp;

		$text_x = $rutex*$rutesx;
		$dot_x = $rutex*$rutesx+$boksbredde-$dot_nudge;
		$linje_x = $rutex*$rutesx+$boksbredde;
	    } else {
		$text_x = $rutex*$rutesx+$text_nudge;
		$dot_x = $rutex*$rutesx+$dot_nudge;
		$linje_x = $rutex*$rutesx;
	    }

	    # Tegn inn teksten
	    $self->{'kartgif'}->string(gdSmallFont,
				       $text_x,
				       ($linjenum+$rutey)*$rutesy,
				       $linje,
				       $self->{'punktfarger'}->{'svart'});

	    # Tegn inn markeringer for hver institusjon i textboksen
	    if ( $forstelinje ) {

		# horisontal linje ved siden av texten
		$self->{'kartgif'}->
		  line($linje_x,
		       ($linjenum+$rutey)*$rutesy+int($rutesy/2)+1,
		       $dot_x,
		       ($linjenum+$rutey)*$rutesy+int($rutesy/2)+1,
		       $self->{'punktfarger'}->{'svart'}); # Br denne vre rd?

		# punkt
		$self->{'kartgif'}->
		  setPixel($dot_x,
			   ($linjenum+$rutey)*$rutesy+int($rutesy/2)+1,
			   gdBrushed);
		$forstelinje = 0;
	    }
	    $linjenum++;
	}
    }

    my $ylinje_start = ($rutey)*$rutesy+int($rutesy/2);
    my $ylinje_slutt = ($linjenum+$rutey-1)*$rutesy+int($rutesy/2)+1;

    my($toppfnutt, $bunnfnutt, $har_bunnfnutt) = (0, 0, 0);

    if ($ypos > ($ylinje_start+$ylinje_slutt)/2) {
	$bunnfnutt = int($rutesy/2);
	$har_bunnfnutt = 1;
    } else {
	$toppfnutt = int($rutesy/2)+1;
    }

    if ((!$hoyre_align && $xpos < $rutex*$rutesx) ||
	($hoyre_align && $xpos > $rutex*$rutesx+$boksbredde)) {

	$bunnfnutt = $toppfnutt = 0;
    }

    $self->{'kartgif'}->
      line($linje_x,
	   $ylinje_start-$toppfnutt+1,
	   $linje_x,
	   $ylinje_slutt+$bunnfnutt,
	   $self->{'punktfarger'}->{'svart'}); # br denne vre rd?

    $self->{'kartgif'}->
      line($linje_x,
	   $har_bunnfnutt?$ylinje_slutt+$bunnfnutt:$ylinje_start-$toppfnutt+1,
	   $xpos,
	   $ypos,
	   $self->{'punktfarger'}->{'svart'}); # br denne vre rd?


    return $self;
}



# -----------------------------------------------------------
#
# Leter ut i fra punktet ($xpos:$ypos) etter ledige ruter p kartet
# som kan romme en $boksx:$boksy stor boks. Bruker bredde-frst sk
# med $retninger som retnings-rekkeflge;
#
# -----------------------------------------------------------
#
# $objekt = $object->_finn_ledig_boks($rutekart, $xpos, $ypos, $boksx, $boksy);
#

sub _finn_ledig_boks {
    my $self = shift;
    my($rutekart, $rutex, $rutey, $antxr, $antyr) = @_;

    my $ant_x = $self->{'ant_x'}; # strrelse p rutenettet
    my $ant_y = $self->{'ant_y'};

    # hvor mange ganger hyde/bredde skal man hoppe unna for  finne
    # et "nytt" sted  sjekke?
    #my $dx_fakt = 0.5 * $antxr;
    #my $dy_fakt = 0.3 * $antyr;

    my $dx_fakt = 1;
    my $dy_fakt = 1;

    # Passe p  ikke sjekke for langt mot hyre kant (noe som er
    # meningslst, siden det allikevel ikke vil vre nok plass der)

    my $startrutex = $rutex;
    my $startrutey = $rutey;

    $startrutex = $ant_x - $antxr + 1 if $rutex > $ant_x - $antxr;
    $startrutey = $ant_y - $antyr + 1 if $rutey > $ant_y - $antyr;

    # koordinat til nrmeste hjrne
    my $xcorner = int(0.5+($rutex/$ant_x)) * $ant_x;
    my $ycorner = int(0.5+($rutex/$ant_y)) * $ant_y;

    # de "andre" hjrnene
    my $other_xcorner = ($xcorner?0:$ant_x);
    my $other_ycorner = ($ycorner?0:$ant_y);

    # retning mot nrmeste hjrne
    my $dx2h = ((2*$rutex-$ant_x)>0?1:-1);
    my $dy2h = ((2*$rutey-$ant_y)>0?1:-1);

    # retning mot motstende hjrne
    my $dx2s = -$dx2h;
    my $dy2s = -$dy2h;

    # Finn ledig plass for text-blokken
    # #######################################
    # #######################################

    # \
    #  sentrum
    #
    #    8 7 5
    #    6 X 3
    #    4 2 1

    #my $retninger = [
#		     [$dx2h, $dy2h],
#		     [0,     $dy2h],
#		     [$dx2h, 0],
#		     [$dx2s, $dy2h],
#		     [$dx2h, $dy2s],
#		     [$dx2s, 0],
#		     [0,     $dy2s],
#		     [$dx2s, $dy2s],
#		    ];

    #$dx2s = $dy2s = 1;
    #$dx2h = $dy2h = -1;

    my $retninger = [
		     [$dx2s, $dy2s],
		     [$dx2s, 0],
		     [$dx2s, $dy2h],
		     [0,     $dy2h],
		     [$dx2h, $dy2h],
		     [$dx2h, 0],
		     [$dx2h, $dy2s],
		     [0,     $dy2s],
		     [$dx2s, $dy2s],
		     [$dx2s, $dy2s],
		    ];

    $retninger = [
		  [$dx2s, 0],
		  [$dx2s, $dy2h],
		  [0,     $dy2h],
		  [$dx2h, $dy2h],
		  [$dx2h, 0],
		  [$dx2h, $dy2s],
		  [0,     $dy2s],
		  [$dx2s, $dy2s],
		  [$dx2s, $dy2s],
		  [$dx2s, $dy2s],
		  [2*$dx2s, $dy2s],
		 ];



    my $sokeliste = [ [$startrutex, $startrutey] ];
    my $been_there = {};

    my $rutesx = $self->{'rutesx'};
    my $rutesy = $self->{'rutesy'};

    my $last_there_rod = [$rutex, $rutey];
    my $last_there_blaa = [$rutex, $rutey];
    $been_there->{$startrutex}->{$startrutey} = 1;

    my $do_lines = 0;

    print STDERR "_finn_ledig_boks($rutekart, $rutex, $rutey, $antxr, $antyr) ($startrutex,$startrutey)\n"
      if $self->{'opt_debug'} >= 2;

    #
    # Om vi nsker  "flge" skerutinen for et bestemt punkt,
    # kommenter ut en av linjene nedenfor, og endre rute-koordinatene
    # til ruta stedet er i.
    #
    #$do_lines = 1 if ($rutex == 41 && $rutey == 47);
    #$do_lines = 1 if ($rutex == 7 && $rutey == 43);

    while (defined($pos = shift @{$sokeliste})) {
	my($sjekkx, $sjekky) = @{$pos};

	# tegne en bl linje for debugging
	if (($last_there_blaa->[0] != $sjekkx || $last_there_blaa->[1] != $sjekky) && $do_lines) {
	    $self->{'kartgif'}->line(($last_there_blaa->[0]*$rutesx)+4,
				     ($last_there_blaa->[1]*$rutesy)+4,
				     ($sjekkx*$rutesx)+5,
				     ($sjekky*$rutesy)+5,
				     $self->{'punktfarger'}->{'signalblaa'});
	    print STDERR "bl [",$last_there_blaa->[0]*$rutesx+4,", ",$last_there_blaa->[1]*$rutesy+4,"] -> [",$sjekkx*$rutesx+4,", ",$sjekky*$rutesx+4,"]\n";
	    $last_there_blaa = [$sjekkx, $sjekky];
	}

	# frst sjekker vi om det er plass der vi str.

	if ($self->_ledig_plass_p($rutekart,
				  $sjekkx, $sjekky,
				  $antxr, $antyr)) {
	    return($sjekkx, $sjekky);
	}

	goto EOJ;

	# s justerer vi litt, og ser om vi fr plass
	my $jsjekkx;
	my $jsjekky;

	if ( (2*$sjekkx-$ant_x)>0 ) {
	    $jsjekkx = $sjekkx - $antxr + $rutekart->[$sjekkx]->[$sjekky]->[0];
	    $jsjekkx = $sjekkx if $jsjekkx < 0;
	} else {
	    my $xtest = $sjekkx;
	    while ($rutekart->[--$xtest]->[$sjekky]->[0]) {
		$jsjekkx = $xtest;
	    }
	}

	#if ($dy2h == 1) { # && (2*$sjekky-$ant_y)>0
	if ( (2*$sjekky-$ant_y)>0 ) {
	    $jsjekky = $sjekky - $antyr + $rutekart->[$jsjekkx]->[$sjekky]->[1];
	    $jsjekky = $sjekky if $jsjekky < 0;
	} else {
	    my $ytest = $sjekky;
	    while ($rutekart->[$jsjekkx]->[--$ytest]->[1]) {
		$jsjekky = $ytest;
	    }
	}

	if (! $been_there->{$jsjekkx}->{$jsjekky}) {
	    if (($last_there_rod->[0] != $jsjekkx || $last_there_rod->[1] != $jsjekky) && $do_lines) {
		$self->{'kartgif'}->line(($last_there_rod->[0]*$rutesx)+4,
					 ($last_there_rod->[1]*$rutesy)+4,
					 ($jsjekkx*$rutesx)+5,
					 ($jsjekky*$rutesy)+5,
					 $self->{'punktfarger'}->{'signalrod'});
		print STDERR "rd [",$last_there_rod->[0]*$rutesx+4,", ",$last_there_rod->[1]*$rutesy+4,"] -> [",$jsjekkx*$rutesx+4,", ",$jsjekky*$rutesx+4,"] $!\n";
		$last_there_rod = [$jsjekkx, $jsjekky];
	    }

	    if ($self->_ledig_plass_p($rutekart,
				      $jsjekkx, $jsjekky,
				      $antxr, $antyr)) {
		return($jsjekkx, $jsjekky);
	    } else {
		$been_there->{$jsjekkx}->{$jsjekky} = 1;
	    }
	}

      EOJ:

	# Hvis vi kom hit, s har vi ikke funnet noe, og m bestemme
	# oss for hvor vi skal sjekke ved neste iterasjon

	foreach my $retning (@{$retninger}) {

	    my $nyrutex = int($sjekkx + ($retning->[0] * $dx_fakt) + 0.5);
	    my $nyrutey = int($sjekky + ($retning->[1] * $dy_fakt) + 0.5);

	    #$nyrutex ||= $sjekkx; # passe p at vi har orntli' verdier
	    #$nyrutey ||= $sjekky;

	    if (! defined($been_there->{$nyrutex}->{$nyrutey}) &&
		$nyrutex >= 0 && $nyrutex <= $ant_x-1 &&
		$nyrutey >= 0 && $nyrutey <= $ant_y-1 ) {

		push(@{$sokeliste}, [$nyrutex, $nyrutey]);
		$been_there->{$nyrutex}->{$nyrutey} = 1;
	    }
	}
    }

    return ($rutex,$rutey);
}




#
# Tegne inn et rutenett p gif-bildet
#
# $object = $object->_tegn_rutenett()
sub _tegn_rutenett {
    my $self = shift;
    my $ant_x = $self->{'ant_x'};
    my $ant_y = $self->{'ant_y'};
    my $rutesx = $self->{'rutesx'};
    my $rutesy = $self->{'rutesy'};

    foreach my $rx (0 .. $ant_x) {
	foreach my $ry (0 .. $ant_y) {
	    $self->{'kartgif'}->line(1,                $ry*$rutesy-1,
				     $self->{'gif_x'}, $ry*$rutesy-1,
				     $self->{'punktfarger'}->{'lysegraa'});
	    $self->{'kartgif'}->line($rx*$rutesx-1,    1,
				     $rx*$rutesx-1,    $self->{'gif_y'},
				     $self->{'punktfarger'}->{'lysegraa'});
	}
    }
    return $self;
}

# -----------------------------------------------------------
#
# Finner ut hvilke typer linjer som skal tegnes, og
# tegner dem ved hjelp av linjer eller piler
#
# -----------------------------------------------------------
#
# $object = $object->tegn_Linje();
#
sub tegn_Linje {
    my($self) = shift;

    my($xpos1,$ypos1,$xpos2,$ypos2,$mpktx,$mpkty);
    my($pilx1,$pilx2,$pily1,$pily2);
    my($sted1,$sted2);
    my($farge1,$farge2);
    my($linjenavn,$linjetype,$ptykk);
    my($tykk, %linjeTykk);
    my(@farge);

    # Definering av ulike farger
    my %linjefarger =
	(
	 "linje",      $self->{'kartgif'}->colorAllocate( 56, 196, 132),
	 "aksess",     $self->{'kartgif'}->colorAllocate(204, 127, 204),
	 "intern",     $self->{'kartgif'}->colorAllocate(  0, 204, 153),
	 "stamnett",   $self->{'kartgif'}->colorAllocate( 51,  51, 204),
	 "knutepunkt", $self->{'kartgif'}->colorAllocate(153, 178, 255)
	);

    my @defaultLinjefarge = (150, 150, 150);
    my $defaultLinjefarge =
	$self->{'kartgif'}->colorAllocate(@defaultLinjefarge);

    # Beregn avstand mellom steder, for  kunne tegne de lengste linjene frst

    $self->beregn_Linje_Lengde();

  linje:
    foreach $linjenavn (reverse sort sortlines keys %{$self->{'_linjelengde'}}){
	print STDERR "Linje : $linjenavn\n" if $self->{'opt_debug'} >= 1;
	if ($self->{'_linjelengde'}->{$linjenavn} <= 0){
	    printf STDERR "Kartgif feil : linjelengde $self->{'_linjelengde'}->{$linjenavn} for $linjenavn\n" if $self->{'opt_v'};
	    next;
	}

	my($sted1) = $self->{'linjedata'}->{$linjenavn}->{'sted1'};
	my($sted2) = $self->{'linjedata'}->{$linjenavn}->{'sted2'};

	# dra inn linjer som henger ut
	($xpos1, $ypos1, $xpos2, $ypos2) = $self->drainn_linje($sted1, $sted2);


        # flytt p parallelle samband
        # prinsipp : parallellforskyvning vinkelrett p
        # - ref likedannede trekanter

	$ptykk = $self->{'pkt_storrelse'} - 2;
	# fast verdi slik at det blir likt med kartmap

	my $lengde = $self->{'_linjelengde'}->{$linjenavn};
	my $stedsted = join(",", sort($sted1, $sted2));

	if (($lengde > 0) && ($tykk = $linjeTykk{$stedsted})){ # et er tegnet allerede
	    my $dt = ($ptykk + 1) / 2 + $tykk;
	    my $dx = $xpos2 - $xpos1;
	    my $dy = $ypos2 - $ypos1;
	    my $ddx = int($dy * $dt / $lengde);
	    my $ddy = int($dx * $dt / $lengde);
	    $xpos1 += $ddx;
	    $ypos1 -= $ddy;
	    $xpos2 += $ddx;
	    $ypos2 -= $ddy;
	}
	$linjeTykk{$stedsted} += $ptykk + 3;

	# Sjekk linjetype, og tegn type linje deretter

	$ptykk = $self->{'linjedata'}->{$linjenavn}->{'ptykk'} ||
	    $self->{'defaultLinjeTykkelse'};
	$linjetype = $self->{'linjedata'}->{$linjenavn}->{'linjetype'};

	# tegn last hvis lastdata ellers ukjent=blue

	if ($self->{'last'} && $self->{'linjedata'}->{$linjenavn}->{'farge'}){
	    if ($self->{'linjedata'}->{$linjenavn}->{'farge'}){
		@farge = @{$self->{'linjedata'}->{$linjenavn}->{'farge'}};
	    } else {
		@farge = @defaultLinjefarge;
	    }
	    #
	    # tegner last ut fra et punkt
	    #
	    my $fargeUt = $self->sett_Farge($farge[3],$farge[4],$farge[5]);
	    $self->tegn_Pil2($linjenavn,
			     $xpos1, $ypos1,
			     $xpos2, $ypos2,
			     $ptykk,
			     $fargeUt);
	    my $fargeInn = $self->sett_Farge($farge[0],$farge[1],$farge[2]);
	    $self->tegn_Pil2($linjenavn,
			     $xpos2, $ypos2,
			     $xpos1, $ypos1,
			     $ptykk,
			     $fargeInn);
	} else {
	    my $farge;
	    if ($self->{'last'}){ # ingen statistikk
		$farge = $defaultLinjefarge;
	    } else {
		$farge = $linjefarger{$linjetype} || $defaultLinjefarge;
	    }

	    $self->tegn_Rett_Linje($xpos1, $ypos1,
				   $xpos2, $ypos2,
				   $ptykk, $farge);
	}
	print STDERR "Linje tegnet: $linjenavn\n" if $self->{'opt_debug'} >= 1;
    }

    return $self;
} # slutt tegn_Linje

# -----------------------------------------------------------
#
# Beregner lengden p linjen som skal tegnes opp mellom
# stedene, for  kunne tegne de lengste linjene frst.
#
# -----------------------------------------------------------
#
# $object = $object->beregn_Linje_Lengde();
#
sub beregn_Linje_Lengde {
    my($self) = shift;
    my($dx, $dy, $lengde, $linjenavn);

    foreach $linjenavn (@{$self->{'_linjenavn'}}) {
	my $sted1 = $self->{'linjedata'}->{$linjenavn}->{'sted1'};
	my $sted2 = $self->{'linjedata'}->{$linjenavn}->{'sted2'};

	my $xpos1 = $self->{'punktdata'}->{$sted1}->{'xpos'};
	my $ypos1 = $self->{'punktdata'}->{$sted1}->{'ypos'};
	my $xpos2 = $self->{'punktdata'}->{$sted2}->{'xpos'};
	my $ypos2 = $self->{'punktdata'}->{$sted2}->{'ypos'};

	$dx = abs($xpos2 - $xpos1);
	$dy = abs($ypos2 - $ypos1);

	$lengde = sqrt(($dx*$dx)+($dy*$dy));
	$self->{'_linjelengde'}->{$linjenavn} = $lengde;

    }
}

# -----------------------------------------------------------
#
# Tegn linje mellom to punkter
#
# -----------------------------------------------------------

sub tegn_Rett_Linje {
    my($self) = shift;

    my($dx, $dy);
    my($lns, $ln);
    my($xpos1, $ypos1, $xpos2, $ypos2, $ptykk, $farge) = @_;

    # dx - delta x, dy - delta y

    $dx = abs($xpos2 - $xpos1);
    $dy = abs($ypos2 - $ypos1);

    # lns - linje senter, ln - linje nummer

    $lns = $ptykk / 2;
    $lns = int ($lns);

    if ($dx < $dy) {
	for($ln=0; $ln<=$ptykk; $ln++) {
	    $self->{'kartgif'}->line($xpos1+$ln-$lns,
				     $ypos1,
				     $xpos2+$ln-$lns,
				     $ypos2,
				     $farge);
	}
    } else {
	for($ln=0; $ln<=$ptykk; $ln++) {
	    $self->{'kartgif'}->line($xpos1,
				     $ypos1+$ln-$lns,
				     $xpos2,
				     $ypos2+$ln-$lns,
				     $farge);
	}
    }
    return $self;
} # slutt tegn_Rett_Linje

# -----------------------------------------------------------
#
# Tegn pil fra en punkt, og halveis til ett annet, dvs.
# at det m tegnes to piler mellom hvert punkt
#
# -----------------------------------------------------------
sub tegn_Pil2 {
    my($self) = shift;
    my($linjenavn, $xpos1, $ypos1, $xpos2, $ypos2, $ptykk, $farge) = @_;

    my($dx,$dy, $ddx, $ddy);
    my($lns, $poly);
    my($ppktx, $ppkty, $mpkty, $mpktx);

    if ($self->{'_linjelengde'}->{$linjenavn} <= 0){
	printf STDERR "Kartgif feil : linjelengde $self->{'_linjelengde'}->{$linjenavn} for $linjenavn\n";
	return;
    }
    $dx = $xpos2 - $xpos1;
    $dy = $ypos2 - $ypos1;

    my $dt = ($ptykk + 1) / 2;
    $ddx = int($dy * $dt / $self->{'_linjelengde'}->{$linjenavn});
    $ddy = int($dx * $dt / $self->{'_linjelengde'}->{$linjenavn});

    if (! defined($self->{'midpktx'}->{$linjenavn})){# for  treffe motsatt pil
	$mpktx = $self->{'midpktx'}->{$linjenavn} = $xpos1 + int($dx / 2);
	$mpkty = $self->{'midpkty'}->{$linjenavn} = $ypos1 + int($dy / 2);
    } else {
	$mpktx = $self->{'midpktx'}->{$linjenavn};
	$mpkty = $self->{'midpkty'}->{$linjenavn};
    }
    $ppktx = $mpktx - int($ptykk * $dx / $self->{'_linjelengde'}->{$linjenavn});
    $ppkty = $mpkty - int($ptykk * $dy / $self->{'_linjelengde'}->{$linjenavn});

    $poly = new GD::Polygon;
    $poly->addPt($xpos1 + $ddx,$ypos1 - $ddy);
    $poly->addPt($ppktx + $ddx, $ppkty - $ddy);
    $poly->addPt($mpktx, $mpkty);
    $poly->addPt($ppktx - $ddx, $ppkty + $ddy);
    $poly->addPt($xpos1 - $ddx, $ypos1 + $ddy);
    $self->{'kartgif'}->filledPolygon($poly, $farge);
}

# -----------------------------------------------------------
#
# Flytt inn linjer som henger utenfor kanten GD kan g i lkke hvis
# punkter tegnes utenfor
#
# -----------------------------------------------------------
#
# ($xpos1, $ypos1, $xpos2, $ypos2) = $object->drainn_linje($sted1, $sted2);
#
sub drainn_linje {
    my($self) = shift;
    my($sted1, $sted2) = @_;

    my(%andre, %ko, $sted);
    $andre{$sted1} = $sted2;
    $andre{$sted2} = $sted1;
    $ko{$sted1} = "$self->{'punktdata'}->{$sted1}->{'xpos'},$self->{'punktdata'}->{$sted1}->{'ypos'}";
    $ko{$sted2} = "$self->{'punktdata'}->{$sted2}->{'xpos'},$self->{'punktdata'}->{$sted2}->{'ypos'}";

    # sjekk hver ende av linja om den er utenfor

    foreach $sted ($sted1, $sted2) {
	if (! defined($self->{'punkt'}->{$sted})) {
	    print STDERR "Ukjent endepunkt $sted\n" if $self->{'opt_v'};
	    next linje;
	}

	my($nyx, $nyy);

	# sjekk om punkt ligger innenfor - trekk det inn

	my($xpos, $ypos) = split(/,/, $ko{$sted});
	my($xpos2, $ypos2) = split(/,/, $ko{$andre{$sted}});
	my($dx) = $xpos - $xpos2;
	my($dy) = $ypos - $ypos2;

	if ($xpos < $self->{'sky_storrelse'}){
	    $nyx = $self->{'sky_storrelse'};
	    $ypos += $dy * ($nyx - $xpos) / $dx if $dx > 0;
	    $xpos = $nyx;
	    $self->{'punktdata'}->{$sted}->{'stedtype'}="satelitt";
	} elsif ($xpos + $self->{'sky_storrelse'} > $self->{'gif_x'}){
	    $nyx = $self->{'gif_x'} - $self->{'sky_storrelse'};
	    $ypos -= $dy * ($xpos - $nyx) / $dx if $dx > 0 ;
	    $xpos = $nyx;
	    $self->{'punktdata'}->{$sted}->{'stedtype'}="satelitt";
	}

	$dx = $xpos - $xpos2;
	$dy = $ypos - $ypos2;

	if ($ypos < $self->{'sky_storrelse'}){
	    $nyy = $self->{'sky_storrelse'};
	    $nyx += $dx * ($nyy - $ypos) / $dy if $dy > 0;
	    $ypos = $nyy;
	    $self->{'punktdata'}->{$sted}->{'stedtype'}="satelitt";
	} elsif ($ypos + $self->{'sky_storrelse'} > $self->{'gif_y'}){
	    $nyy = $self->{'gif_y'} - $self->{'sky_storrelse'};
	    $nyx -= $dx * ($ypos - $nyy) / $dy if $dy > 0;
	    $ypos = $nyy;
	    $self->{'punktdata'}->{$sted}->{'stedtype'}="satelitt";
	}

	$ko{$sted} = "$xpos,$ypos";  # for bruk i linjetegning
	$self->{'punktdata'}->{$sted}->{'xpos'} = $xpos; # for punkttegning
	$self->{'punktdata'}->{$sted}->{'ypos'} = $ypos;
    }

    return(split(",",$ko{$sted1}), split(",",$ko{$sted2}));
} # slutt drainn_linje

# -----------------------------------------------------------
#
# returnerer mtime-informasjoen om en fil.
#
# -----------------------------------------------------------
sub _mtime {
    my($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size,
       $atime, $mtime, $ctime, $blksize, $blocks)
	= stat(shift);
    return $mtime;
}

#
# IKKE I BRUK
#
#sub utenfor {
#    my($self) = shift;
#    my($x, $y)=@_;
#    return($y < 0 || $y > $self->{'gif_y'} ||
#	   $x < 0 || $x > $self->{'gif_x'} || 0);
#}

#
# IKKE I BRUK
#
#sub move_inside {
#    my($self) = shift;
#    my($p) = shift;
#    my ($x, $y) = ($self->{'punktdata'}->{$p}->{'xpos'},
#		   $self->{'punktdata'}->{$p}->{'ypos'});
#    if ($y < 0){
#	$y = $self->{'sky_storrelse'};
#    } elsif ($y > $self->{'gif_y'}){
#	$y = $self->{'gif_y'} - $self->{'sky_storrelse'};
#    }
#
#    if ($x < 0){
#	$x = $self->{'sky_storrelse'};
#    } elsif ($x > $self->{'gif_x'}){
#	$x = $self->{'gif_x'} - $self->{'sky_storrelse'};
#    }
#    ($self->{'punktdata'}->{$p}->{'xpos'}, $self->{'punktdata'}->{$p}->{'ypos'})=($x, $y);
#
# Denne feiler kontrollen - hvor tar vi $navn fra? -- sjn 19990616
####    $self->{'punktdata'}->{$navn}->{'stedtype'} = "satelitt";
#}

# #####################################
#
# sort lines by descending length and size
#
sub sortlines {
    my($self) = shift;
    my($cmp);
    local $^W = 0;

    return $self->{'_linjelengde'}->{$b} <=> $self->{'_linjelengde'}->{$a} ||
      $self->{'linjedata'}->{$b}->{'ptykk'} <=> $self->{'linjedata'}->{$b}->{'ptykk'};
}

# ######################################
#
# $object = $object->_skriv_textpos($sted_id, $xpos, $ypos, $navn);
#
sub _skriv_textpos {
    my $self = shift;
    my($sted, $nyposy, $nyposx, $pnavn) = @_;
    if (defined($self->{'textpos'})) {
	printf STDERR "PUNKT text %-15s gif %3d %3d (%s)\n",
	  $sted, $nyposy, $nyposx, $pnavn;
    }
    return $self;
}



# datere kartet
#
# $object = $object->dater_Kart($unix_tid)

sub dater_Kart {
    my($self) = shift;
    my($unix_time) = shift || 0;

    $unix_time = _mtime($self->{'last'})
	if (defined($self->{'last'}) && -f $self->{'last'});
    $unix_time ||= time; # sette tid til "nuh!" om ikke gitt.

    my($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) =
	localtime($unix_time);
    my $time = sprintf "Data fra %02d-%02d-%02d %02d:%02d",
      1900+$year, $mon+1, $mday, $hour, $min;

    my $xpos = $self->{'gif_x'} - (length($time) * gdSmallFont->width) - 1;
    my $ypos = $self->{'gif_y'} - gdSmallFont->height - 1;
    #my $ypos = 0;

    $self->{'kartgif'}->string(gdSmallFont,
			       $xpos,
			       $ypos,
			       $time,
			       $self->{'punktfarger'}->{'svart'});

# Reservere ruter p kartet, slik at ikke noe skriver over
# daterings-strengen!
#
#    my $rutex = int($xpos*$self->{'rutesx'});
#    my $rutey = int($ypos*$self->{'rutesy'});
#    for (my $rute); $rute >= length($time); $rute++) {
#	$self->{'_sone'}->{$rute}->{0} = 1;
#    }

    return $self;
}

1; # moduler m returnere en sann verdi
